#  2019-09-06 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writup
#
# License:  AGPLv3
# (c) 2019. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the event handler module modifying request parameters.
"""

import logging
import re

from privacyidea.lib import _
from privacyidea.lib.eventhandler.base import BaseEventHandler
from privacyidea.lib.user import User
from privacyidea.lib.utils import is_true

log = logging.getLogger(__name__)


class ACTION_TYPE():
    """
    Allowed actions
    """
    SET = "set"
    DELETE = "delete"


class RequestManglerEventHandler(BaseEventHandler):
    """
    An Eventhandler needs to return a list of actions, which it can handle.

    It also returns a list of allowed action and conditions

    It returns an identifier, which can be used in the event-handling definitions
    """

    identifier = "RequestMangler"
    description = "This event handler can modify the parameters in the request."

    @property
    def allowed_positions(self):
        """
        This returns the allowed positions of the event handler definition.
        :return: list of allowed positions
        """
        """
        Usually we would only modify the parameters in the PRE location, so that
        the request is handled with the modified parameters. 
        
        But we could also modify the parameters in the POST location, so that the request
        is handled with the original parameters, but *after* the request is handled,
        some parameters can be changed to that an event handler, that is called *After*
        the RequestMangler gets other input parameters.
        
        At the time of writing I can not make up a scenario, but technically it could 
        make sense.
        """
        return ["post", "pre"]

    @property
    def actions(self):
        """
        This method returns a dictionary of allowed actions and possible
        options in this handler module.

        :return: dict with actions
        """
        actions = {ACTION_TYPE.DELETE:
                       {"parameter":
                            {"type": "str",
                             "required": True,
                             "description": _("The parameter that should be deleted.")}
                        },
                   ACTION_TYPE.SET:
                       {"parameter":
                            {"type": "str",
                             "required": True,
                             "description": _("The parameter that should be added or modified.")
                             },
                        "value":
                            {"type": "str",
                             "required": True,
                             "description": _("The new value of the parameter. Can contain tags like {0}, {1} for "
                                              "the matched sub strings.")
                             },
                        "match_parameter":
                            {"type": "str",
                             "description": _("The parameter, that should match some values.")
                             },
                        "match_pattern":
                            {"type": "str",
                             "description": _("The value of the match_parameter. It can contain a regular "
                                              "expression and '()' to transfer values to the new parameter.")
                             },
                        "reset_user":
                            {"type": "bool",
                             "description": _("If the parameter is 'username', 'user' or 'realm', the user will be"
                                              " reset. This can have an effect on any further actions on the user!")}
                        }
                   }
        return actions

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action:
        :param options: Contains the flask parameters g, request, response
            and the handler_def configuration
        :type options: dict
        :return:
        """
        ret = True
        request = options.get("request")
        handler_def = options.get("handler_def")
        handler_options = handler_def.get("options", {})
        parameter = handler_options.get("parameter")

        if parameter:
            if action.lower() == ACTION_TYPE.DELETE:
                if parameter in request.all_data:
                    del (request.all_data[parameter])
            elif action.lower() == ACTION_TYPE.SET:
                value = handler_options.get("value")
                match_parameter = handler_options.get("match_parameter")
                match_pattern = handler_options.get("match_pattern")

                if value is not None:
                    # We only take action, if we have a value, even an empty string "".
                    if not match_parameter:
                        # simple setting a parameter
                        request.all_data[parameter] = value
                    elif match_pattern and match_parameter in request.all_data:
                        # setting a parameter depending on another value,
                        # but only set it, if match_parameter exists
                        """
                        Note: Beware user supplied format-string like "match_pattern", can
                        be dangerous: http://lucumr.pocoo.org/2016/12/29/careful-with-str-format/
                        but in our case it is fine because no objects are involved, as m.groups() 
                        always returns a tuple of strings
                        """
                        m = re.match("^" + match_pattern + "$", request.all_data.get(match_parameter))
                        if m:
                            # Now we set the new value with the matching tuple
                            try:
                                new_value = value.format(*m.groups())
                                request.all_data[parameter] = new_value
                                reset_user = is_true(handler_options.get("reset_user"))

                                # Optionally reset the user if a param of the user was mangled
                                # TODO this should be a UserMangler to explicitly change the user
                                # TODO then remove any user info from all_data...
                                if parameter in ["realm", "username", "user"] and reset_user:
                                    realm = request.all_data.get("realm")
                                    user = request.all_data.get("username") or request.all_data.get("user")
                                    user = User(login=user, realm=realm)
                                    request.User = user
                            except IndexError:
                                log.warning(f"The number of found tags ({m.groups()!r}) "
                                            f"do not match the required number ({value!r}).")

        return ret



#  2020-09-24 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Use Argon2
#  2017-08-11 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             initial writeup
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNE7SS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from ..models import AuthCache, db
from sqlalchemy import and_
from passlib.hash import argon2
import datetime
import logging

ROUNDS = 9
log = logging.getLogger(__name__)


def _hash_password(password):
    return argon2.using(rounds=ROUNDS).hash(password)


def add_to_cache(username, realm, resolver, password):
    # Can not store timezone aware timestamps!
    first_auth = datetime.datetime.utcnow()
    auth_hash = _hash_password(password)
    record = AuthCache(username, realm, resolver, auth_hash, first_auth, first_auth)
    log.debug('Adding record to auth cache: ({!r}, {!r}, {!r}, {!r})'.format(
        username, realm, resolver, auth_hash))
    r = record.save()
    return r


def update_cache(cache_id):
    last_auth = datetime.datetime.utcnow()
    db.session.query(AuthCache).filter(
        AuthCache.id == cache_id).update({"last_auth": last_auth,
                                          AuthCache.auth_count: AuthCache.auth_count + 1})
    db.session.commit()


def delete_from_cache(username, realm, resolver, password, last_valid_cache_time=None, max_auths=0):
    """
    Deletes all authcache entries that match the user and either match the password, are expired, or have reached the
    maximum number of allowed authentications.

    :param username:
    :param realm:
    :param resolver:
    :param password:
    :param last_valid_cache_time: Oldest valid time for a cache entry to be still valid. I.e., if the first
    authentication of the entry is before this time point, it is not valid anymore.
    :param max_auths: Maximum number of allowed authentications.
    """
    cached_auths = db.session.query(AuthCache).filter(AuthCache.username == username,
                                                      AuthCache.realm == realm,
                                                      AuthCache.resolver == resolver).all()
    r = 0
    for cached_auth in cached_auths:
        delete_entry = False
        # if the password matches or the entry is otherwise invalid, we deleted it.
        try:
            if max_auths > 0 and cached_auth.auth_count >= max_auths:
                delete_entry = True
            elif last_valid_cache_time and cached_auth.first_auth < last_valid_cache_time:
                delete_entry = True
            elif argon2.verify(password, cached_auth.authentication):
                delete_entry = True

        except ValueError:
            log.debug("Old (non-argon2) authcache entry for user {0!s}@{1!s}.".format(username, realm))
            # Also delete old entries
            delete_entry = True
        if delete_entry:
            r += 1
            cached_auth.delete()
    db.session.commit()
    return r


def cleanup(minutes):
    """
    Will delete all authcache entries, where last_auth column is older than
    the given minutes.

    :param minutes: Age of the last_authentication in minutes
    :type minutes: int
    :return:
    """
    cleanuptime = datetime.datetime.utcnow() - datetime.timedelta(minutes=minutes)
    r = db.session.query(AuthCache).filter(AuthCache.last_auth < cleanuptime).delete()
    db.session.commit()
    return r


def verify_in_cache(username, realm, resolver, password, first_auth=None, last_auth=None,
                    max_auths=0):
    """
    Verify if the given credentials are cached and if the time is correct.
    
    :param username: 
    :param realm: 
    :param resolver: 
    :param password: 
    :param first_auth: The timestamp when the entry was first written to the 
        cache. Only find newer entries 
    :param last_auth: The timestamp when the entry was last successfully 
        verified. Only find newer entries
    :param max_auths: Maximum number of times the authcache entry can be used to skip
        authentication, as defined by ACTION.AUTH_CACHE policy. Will return False if the current number of
        authentications + 1 of the cached authentication exceeds this value.
    :type max_auths: int
    :return: 
    """
    conditions = []
    result = False
    conditions.append(AuthCache.username == username)
    conditions.append(AuthCache.realm == realm)
    conditions.append(AuthCache.resolver == resolver)

    if first_auth:
        conditions.append(AuthCache.first_auth > first_auth)
    if last_auth:
        conditions.append(AuthCache.last_auth > last_auth)

    filter_condition = and_(*conditions)
    cached_auths = AuthCache.query.filter(filter_condition).all()

    for cached_auth in cached_auths:
        try:
            result = argon2.verify(password, cached_auth.authentication)
        except ValueError:
            log.debug("Old (non-argon2) authcache entry for user {0!s}@{1!s}.".format(username, realm))
            result = False

        if result and max_auths > 0:
            # Check if auth_count allows this authentication too
            result = cached_auth.auth_count < max_auths

        if result:
            # Update the last_auth and the auth_count
            update_cache(cached_auth.id)
            break

    if not result:
        # Delete older entries
        delete_from_cache(username, realm, resolver, password, first_auth, max_auths)

    return result



# SPDX-FileCopyrightText: (C) 2024 Paul Lettich <paul.lettich@netknights.it>
# SPDX-FileCopyrightText: (C) 2015 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Test the models of the privacyIDEA database."""
from mock import mock
import os
from sqlalchemy import func

from privacyidea.models import (Token,
                                Resolver,
                                ResolverRealm, NodeName,
                                TokenRealm,
                                ResolverConfig,
                                Realm,
                                Config,
                                Policy,
                                Challenge, MachineResolver,
                                MachineResolverConfig, MachineToken, Admin,
                                CAConnector, CAConnectorConfig, SMTPServer,
                                PasswordReset, EventHandlerOption,
                                EventHandler, SMSGateway,
                                EventHandlerCondition, PrivacyIDEAServer,
                                ClientApplication, Subscription, UserCache,
                                EventCounter, PeriodicTask, PeriodicTaskLastRun,
                                PeriodicTaskOption, MonitoringStats, PolicyCondition, db,
                                Tokengroup, TokenTokengroup, Serviceid)
from .base import MyTestCase
from dateutil.tz import tzutc
from datetime import datetime
from datetime import timedelta


class TokenModelTestCase(MyTestCase):
    """
    Test the token on the database level
    """
    def create_resolver_realm(self):
        r = Resolver("resolver1", "passwdresolver")
        r.save()
        self.assertTrue(r.name is not None, r.name)
        self.assertTrue(r.rtype is not None, r.rtype)
        # Add configuration to the resolver
        conf = ResolverConfig(r.id, "fileName", "tests/testdata/passwd")
        conf.save()

        # Read Resolver
        r1 = Resolver.query.filter_by(name="resolver1").first()
        self.assertTrue(r1.rtype == "passwdresolver", r1.rtype)

        realm = Realm("realm1")
        realm.save()
        # Put the resolver into the realm
        rrealm = ResolverRealm(resolver_name="resolver1",
                               realm_name="realm1")
        rrealm.save()

    def test_00_create_token(self):
        otpkey = "1234567890"
        t1 = Token(serial="serial1",
                   otpkey=otpkey,
                   tokentype="hmac")
        t1.set_description("myfirsttoken")
        t1.set_hashed_pin("1234")
        t1.otplen = 6
        tid = t1.save()

        userpin = "HalloDuda"
        t1.set_user_pin(userpin)
        t1.save()

        pin_object = t1.get_user_pin()
        self.assertTrue(pin_object.getPin() == userpin.encode('utf8'))

        t = Token.query.filter_by(id=tid).first()
        self.assertTrue(len(t.pin_hash) > 0)
        self.assertTrue(len(t.user_pin) > 0)

        otp_obj = t.get_otpkey()
        self.assertTrue(otp_obj.getKey() == otpkey.encode('utf8'))
        count = t.count
        self.assertTrue(count == 0)

        up = t.get_user_pin()
        self.assertTrue(up.getPin() == userpin.encode('utf8'))

        self.assertTrue(t.check_pin("1234"))

        t.set_user_pin(b'HalloDuDa')
        self.assertTrue(t.get_user_pin().getPin() == b'HalloDuDa')

        t.set_user_pin('HelloWörld')
        self.assertTrue(t.get_user_pin().getPin().decode('utf8') == 'HelloWörld')

        t.set_hashed_pin(b'1234')
        self.assertTrue(t.check_pin(b'1234'))

        t.set_hashed_pin('HelloWörld')
        self.assertTrue(t.check_pin('HelloWörld'))

        t.pin_hash = None
        self.assertTrue(t.check_pin(''))
        self.assertFalse(t.check_pin(None))
        self.assertFalse(t.check_pin('1234'))

        t.pin_hash = ''
        self.assertTrue(t.check_pin(''))
        self.assertFalse(t.check_pin('1234'))

        t.set_hashed_pin('')
        self.assertTrue(len(t.pin_hash) > 0)
        self.assertTrue(t.check_pin(''))
        self.assertFalse(t.check_pin('1234'))

        # Delete the token
        t1.delete()
        t = Token.query.filter_by(id=tid).first()
        self.assertTrue(t is None)

    def test_01_create_a_token_with_a_realm(self):
        """
        Create a token with a user and a tokenrealm in the database
        When we create a token with a user, the tokenrealm is filled in
        automatically.
        """
        self.create_resolver_realm()
        # Now we have a user cornelius@realm1
        # userid=1009
        # resolver=resolver1
        # realm=realm1
        otpkey = "123456"

        # create token and also assign the user and realm
        Token(serial="serial2",
              otpkey=otpkey,
              userid=1009,
              resolver="resolver1",
              realm="realm1")
        t2 = Token.query\
                  .filter_by(serial="serial2")\
                  .first()
        self.assertEqual(t2.first_owner.resolver, "resolver1")
        # check the realm list of the token
        realm_found = False
        for realm_entry in t2.realm_list:
            if realm_entry.realm.name == "realm1":
                realm_found = True
        self.assertTrue(realm_found)

        # set the pin in a hashed way
        t2.set_pin("thepin")
        t2.save()
        r = t2.check_pin("wrongpin")
        self.assertFalse(r)
        r = t2.check_pin("thepin")
        self.assertTrue(r)

        # save the pin in an encrypted way
        t2.set_pin("thepin", hashed=False)
        t2.save()
        r = t2.check_pin("wrongpin")
        self.assertTrue(r is False)
        r = t2.check_pin("thepin")
        self.assertTrue(r)
        pin = t2.get_pin()
        self.assertEqual("thepin", pin)
        t2.set_pin('pinwithä', hashed=False)
        self.assertEqual(t2.get_pin(), 'pinwithä')

        # set the so pin
        (enc, iv) = t2.set_so_pin("topsecret")
        self.assertTrue(len(enc) > 0)
        self.assertTrue(len(iv) > 0)

        (enc, iv) = t2.set_so_pin(b"topsecret")
        self.assertTrue(enc)
        self.assertTrue(iv)

        (enc, iv) = t2.set_so_pin("topsecrät")
        self.assertTrue(enc)
        self.assertTrue(iv)

        # get token information
        token_dict = t2.get()
        self.assertIsInstance(token_dict, dict)
        self.assertTrue(token_dict.get("resolver") == "resolver1")
        # THe realm list is contained in realms
        self.assertTrue("realm1" in token_dict.get("realms"))

        resolver = t2.get("resolver")
        self.assertTrue(resolver == "resolver1")

        # retrieve a value, that does not exist
        nonexist = t2.get("nonexist")
        self.assertTrue(nonexist is None)

        # setting normal values
        t2.count_window = 100
        t2.otplen = 8
        t2.set_description("De scription")
        t2.save()
        t2.set_info({"info": "value"})
        t3 = Token.query\
                  .filter_by(serial="serial2")\
                  .first()
        self.assertTrue(t3.count_window == 100)
        self.assertTrue(t3.otplen == 8)
        self.assertTrue(t3.description == "De scription")
        self.assertTrue(t3.info_list[0].Value == "value")
        self.assertTrue(t3.get_info().get("info") == "value")

        # test the string representation
        s = "{0!s}".format(t3)
        self.assertEqual("serial2", s)

        # update token type
        t2.update_type("totp")
        self.assertTrue(t2.get("tokentype") == "totp")

        old_key = t2.key_enc
        old_pin = t2.pin_hash
        t2.update_token(description="New Text", otpkey="1234",
                        pin="new pin")
        self.assertTrue(t2.get("description") == "New Text")
        # The key has changed
        self.assertTrue(t2.key_enc != old_key)
        self.assertTrue(t2.pin_hash != old_pin)

        t2.set_otpkey(b'12345')
        self.assertEqual(b'12345', t2.get_otpkey().getKey(), t2)
        t2.failcount = 5
        t2.set_otpkey('Hellö', reset_failcount=True)
        self.assertTrue(t2.failcount == 0, t2)
        self.assertEqual('Hellö', t2.get_otpkey().getKey().decode('utf8'), t2)

        # key too long
        k = os.urandom(1500)
        t2.set_otpkey(k)
        self.assertGreater(len(t2.key_enc), Token.key_enc.property.columns[0].type.length)

        # SQLite supports writing too long data, all others don't.
        if db.engine.name != 'sqlite':
            self.assertRaises(Exception, db.session.commit)
            db.session.rollback()

        # set an empty token description
        self.assertEqual(t2.set_description(desc=None), '')

        # delete the token
        ret = t2.delete()
        self.assertTrue(ret)
        # check that the TokenRealm is deleted
        q = TokenRealm.query.all()
        self.assertTrue(len(q) == 0)

    def test_02_config_model(self):
        c = Config("splitRealm", True,
                   Type="string", Description="something")

        cid = c.save()
        self.assertTrue(cid == "splitRealm", cid)
        self.assertTrue("{0!s}".format(c) == "<splitRealm (string)>", c)

        # delete the config
        config = Config.query.filter_by(Key="splitRealm").first()
        self.assertTrue(config.delete() == "splitRealm")
        q = Config.query.filter_by(Key="splitRealm").all()
        # the list is empty
        self.assertTrue(len(q) == 0, q)

    def test_03_create_and_delete_realm(self):
        realmname = "realm23"
        r = Realm(realmname)
        r.save()

        qr = Realm.query.filter_by(name=realmname).first()
        self.assertTrue(qr.name == realmname)

        # delete realm
        qr.delete()
        # no realm left
        q = Realm.query.filter_by(name=realmname).all()
        self.assertTrue(len(q) == 0)

    def test_04_update_resolver_config(self):
        resolvername = "resolver2"
        r = Resolver(resolvername, "passwdresolver")
        rid = r.save()
        self.assertTrue(r.name is not None, r.name)
        self.assertTrue(r.rtype == "passwdresolver", r.rtype)
        # save first resolver config
        rc = ResolverConfig(resolver=resolvername,
                            Key="fileName",
                            Value="/etc/passwd")
        res_conf_id = rc.save()
        self.assertTrue(res_conf_id > 0, res_conf_id)
        # update resolver config
        rc = ResolverConfig(resolver=resolvername,
                            Key="fileName",
                            Value="/etc/secureusers")
        res_conf_id2 = rc.save()
        self.assertTrue(res_conf_id2 == res_conf_id,
                        res_conf_id2)
        # delete resolver and its config
        r.delete()
        # check that config is empty
        q = ResolverConfig.query.filter_by(resolver_id=rid).all()
        self.assertTrue(len(q) == 0, q)

    def test_05_get_set_realm(self):
        t1 = Token(serial="serial1123")
        t1.save()
        realms = t1.get_realms()
        self.assertTrue(len(realms) == 0)
        t1.set_realms(["realm1"])
        t1.save()
        realms = t1.get_realms()
        self.assertTrue(len(realms) == 1)

    def test_06_caconnector(self):
        connector_name = "testCA"
        # create a CA connector
        cacon = CAConnector(name=connector_name, catype="localCA")
        cacon.save()

        # try to create a CA connector, that already exist
        # cacon = CAConnector(name="testCA", catype="localCA")
        # self.assertRaises(Exception, cacon.save)

        # add config entries to the CA connector
        CAConnectorConfig(caconnector_id=1, Key="Key1",
                          Value="Value1").save()
        CAConnectorConfig(caconnector=connector_name, Key="Key2",
                          Value="Value2", Type="password").save()
        q = CAConnectorConfig.query.filter_by(caconnector_id=1).all()
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].Value, "Value1")
        self.assertEqual(q[1].Value, "Value2")

        # update config entries
        CAConnectorConfig(caconnector=connector_name, Key="Key2",
                          Value="Value3").save()
        q = CAConnectorConfig.query.filter_by(Key="Key2").all()
        self.assertEqual(len(q), 1)
        self.assertEqual(q[0].Value, "Value3")

        # delete config entries
        CAConnectorConfig.query.filter_by(Key="Key2").delete()
        q = CAConnectorConfig.query.filter_by(Key="Key2").all()
        self.assertEqual(q, [])

        # Delete the CA connector. Remaining Config entries will be deleted
        # automatically
        cacon = CAConnector.query.filter_by(name=connector_name).first()
        r = cacon.delete()
        self.assertEqual(r, 1)
        q = CAConnectorConfig.query.filter_by(Key="Key1").all()
        # FIXME: The last entry does not get deleted!
        # self.assertEqual(q, [])

    def test_10_delete_resolver_realm(self):
        resolvername = "res1"
        realmname = "r1"
        # create a resolver
        r = Resolver(resolvername, "passwdresolver")
        _rid = r.save()
        # create a realm with this resolver
        realm = Realm(realmname)
        _realm_id = realm.save()
        rr = ResolverRealm(realm_name=realmname,
                           resolver_name=resolvername)
        rr_id = rr.save()
        self.assertTrue(rr_id > 0, rr_id)
        # check how many resolvers are in the realm
        db_realm = Realm.query.filter_by(name=realmname).first()
        self.assertTrue(len(db_realm.resolver_list) == 1,
                        len(db_realm.resolver_list))
        # remove the resolver from the realm
        # we can do this by deleting rr_id
        rr.delete()
        # check how many resolvers are in the realm
        self.assertTrue(len(db_realm.resolver_list) == 0,
                        len(db_realm.resolver_list))
        # delete the realm
        db_realm.delete()

    def test_11_policy(self):
        p = Policy("pol1", active="true",
                   scope="selfservice", action="action1",
                   realm="*")
        p.save()
        self.assertTrue(p.action == "action1", p)
        self.assertTrue("action1" in p.get().get("action"), p)
        self.assertTrue("action1" in p.get("action"), p)
        self.assertEqual(p.get()["conditions"], [])

        p2 = Policy("pol1", active="false",
                    scope="selfservice", action="action1",
                    realm="*")
        self.assertFalse(p2.active, p2.active)

        # update
        self.assertTrue(p.user == "", p.user)
        p.user = "cornelius"
        p.resolver = "*"
        p.client = "0.0.0.0"
        p.time = "anytime"
        p.pinode = "pinode1, pinode2"
        p.save()
        self.assertTrue(p.user == "cornelius", p.user)
        self.assertEqual(p.pinode, "pinode1, pinode2")

        # save admin policy
        p3 = Policy("pol3", active="false", scope="admin",
                    adminrealm='superuser', action="*", pinode="pinode3")
        self.assertEqual(p3.adminrealm, "superuser")
        self.assertEqual(p3.pinode, "pinode3")
        p3.save()

        # set conditions
        p3.set_conditions([("userinfo", "type", "==", "foobar", False),
                           ("request", "user_agent", "==", "abcd", True)])
        self.assertEqual(p3.get_conditions_tuples(),
                         [("userinfo", "type", "==", "foobar", False),
                          ("request", "user_agent", "==", "abcd", True)])
        self.assertEqual(p3.get()["conditions"],
                         [("userinfo", "type", "==", "foobar", False),
                          ("request", "user_agent", "==", "abcd", True)])
        self.assertEqual(PolicyCondition.query.count(), 2)

        p3.set_conditions([("userinfo", "type", "==", "baz", True)])
        p3.save()
        self.assertEqual(p3.get()["conditions"],
                         [("userinfo", "type", "==", "baz", True)])
        self.assertEqual(len(p3.conditions), 1)
        self.assertEqual(p3.conditions[0].Value, "baz")
        self.assertEqual(PolicyCondition.query.count(), 1)

        # Check that the change has been persisted to the database
        p3_reloaded1 = Policy.query.filter_by(name="pol3").one()
        self.assertEqual(p3_reloaded1.get()["pinode"], ["pinode3"])
        self.assertEqual(p3_reloaded1.get()["conditions"],
                         [("userinfo", "type", "==", "baz", True)])
        self.assertEqual(len(p3_reloaded1.conditions), 1)
        self.assertEqual(p3_reloaded1.conditions[0].Value, "baz")
        self.assertEqual(PolicyCondition.query.count(), 1)

        p3.set_conditions([])
        p3.save()
        self.assertEqual(p3.get()["conditions"], [])
        self.assertEqual(Policy.query.filter_by(name="pol3").one().get()["conditions"], [])
        self.assertEqual(PolicyCondition.query.count(), 0)

        # Test policies with adminusers
        p = Policy("pol1admin", active="true",
                   scope="admin", action="action1",
                   adminuser="jan, hein, klaas, pit")
        r = p.save()
        adminusers = p.get("adminuser")
        self.assertEqual(["jan", "hein", "klaas", "pit"], adminusers)
        p2 = Policy.query.filter_by(id=r).one()
        self.assertEqual("jan, hein, klaas, pit", p2.adminuser)

    def test_12_challenge(self):
        c = Challenge("S123456")
        self.assertTrue(len(c.transaction_id) == 20, c.transaction_id)
        self.assertTrue(len(c.get_transaction_id()) == 20, c.transaction_id)

        c.set_data("some data")
        self.assertTrue(c.data == "some data", c.data)
        self.assertTrue(c.get_data() == "some data", c.data)
        c.set_data({"some": "data"})
        self.assertTrue("some" in c.data, c.data)
        c.set_session("session")
        self.assertTrue(c.get_session() == "session", c.session)
        c.set_challenge("challenge")
        self.assertTrue(c.get_challenge() == "challenge", c.challenge)

        self.assertTrue("otp_received" in "{0!s}".format(c), "{0!s}".format(c))
        self.assertTrue("transaction_id" in "{0!s}".format(c), "{0!s}".format(c))
        self.assertTrue("timestamp" in "{0!s}".format(c), "{0!s}".format(c))

        # test with timestamp=True, which results in something like this:
        timestamp = '2014-11-29 21:56:43.057293'
        self.assertTrue(len(c.get(True).get("timestamp")) == len(timestamp),
                        c.get(True))
        # otp_status
        c.set_otp_status(valid=False)
        self.assertTrue(c.get_otp_status()[0], c.get_otp_status())
        self.assertFalse(c.get_otp_status()[1], c.get_otp_status())

    def test_13_machine_resolver(self):
        # create the machineresolver and a config entry
        mr = MachineResolver("mr1", "mrtype1")
        mr_id = mr.save()
        self.assertTrue(mr_id > 0, mr)
        mrc = MachineResolverConfig(resolver="mr1", Key="key1", Value="value1")
        mrc_id = mrc.save()
        self.assertTrue(mrc_id > 0, mrc)
        # check that the config entry exist
        db_mrconf = MachineResolverConfig.query.filter(
            MachineResolverConfig.resolver_id == mr_id).first()
        self.assertTrue(db_mrconf is not None)

        # add a config value by ID
        mrc = MachineResolverConfig(resolver_id=mr_id, Key="key2", Value="v2")
        mrc_id = mrc.save()
        self.assertTrue(mrc_id > 0)
        # update config
        MachineResolverConfig(resolver_id=mr_id, Key="key2",
                              Value="new value").save()
        # check if the value is updated.
        new_config = MachineResolverConfig.query.filter(
            MachineResolverConfig.Key == "key2").first()
        self.assertTrue(new_config.Value == "new value", new_config.Value)

        # Connect a machine to a token
        mt_id = MachineToken(machineresolver_id=mr_id, machine_id="client1",
                             serial="serial1123",
                             application="SSH").save()
        self.assertTrue(mt_id > 0, mt_id)
        # Connect another machine to a token
        token_id = Token.query.filter_by(serial="serial1123").first().id
        mt_id2 = MachineToken(machineresolver="mr1", machine_id="client2",
                              token_id=token_id,
                              application="LUKS").save()
        self.assertTrue(mt_id2 > mt_id, (mt_id2, mt_id))
        # get the token that contains the machines
        db_token = Token.query.filter_by(serial="serial1123").first()
        # check the length of the machine list of the token
        self.assertTrue(len(db_token.machine_list) == 2, db_token.machine_list)
        machine2 = db_token.machine_list[1].machine_id
        self.assertTrue(machine2 == "client2", (machine2,
                                                db_token.machine_list))

        # delete the machine resolver
        db_mr = MachineResolver.query.filter(MachineResolver.name ==
                                             "mr1").first()
        db_mr.delete()
        # check that there is no machine resolver and no config entry

        db_mr = MachineResolver.query.filter(MachineResolver.name ==
                                             "mr1").first()
        self.assertTrue(db_mr is None)
        db_mrconf = MachineResolverConfig.query.filter(
            MachineResolverConfig.resolver_id == mr_id).first()
        self.assertTrue(db_mrconf is None)

    def test_14_save_update_admin(self):
        # create an admin user
        adminname = Admin(username="admin", password="secret",
                          email="admin@privacyidea.org").save()
        self.assertEqual(adminname, "admin")
        password1 = Admin.query.filter_by(username="admin").first().password

        # update admin - change the password
        Admin(username="admin", password="supersecret").save()
        password2 = Admin.query.filter_by(username="admin").first().password
        self.assertTrue(password1 != password2, (password1, password2))

    def test_15_add_and_delete_tokeninfo(self):
        t1 = Token("serialTI")
        t1.save()

        t1.set_info({"key1": "value1",
                     "key2": "value2",
                     "key3": "value3"})
        t2 = Token.query.filter_by(serial="serialTI").first()
        t2info = t2.get_info()
        self.assertTrue(t2info.get("key2") == "value2", t2info)

        t2.del_info("key2")
        t2info = t2.get_info()
        self.assertTrue(t2info.get("key2") is None, t2info)

    def test_16_add_and_delete_tokeninfo_password(self):
        t1 = Token("serialTI2")
        t1.set_info({"key1": "value1",
                     "key1.type": "password"})

        t2 = Token.query.filter_by(serial="serialTI2").first()
        t2info = t2.get_info()

        self.assertTrue(t2info.get("key1.type") == "password",
                        t2info)

    def test_17_add_and_delete_smtpserver(self):
        s1 = SMTPServer(identifier="myserver", server="1.2.3.4")
        s1.save()
        s2 = SMTPServer.query.filter_by(identifier="myserver").first()
        self.assertTrue(s2.server, "1.2.3.4")

        # Update the server
        SMTPServer(identifier="myserver", server="100.2.3.4",
                   username="user", password="password", tls=True,
                   description="test", port=123).save()
        modified_server = SMTPServer.query.filter_by(
            identifier="myserver").first()

        self.assertEqual(modified_server.server, "100.2.3.4")
        self.assertEqual(modified_server.username, "user")

        # Delete the server
        s2.delete()
        # Try to find the server
        s2 = SMTPServer.query.filter_by(identifier="myserver").first()
        # The server does not exist anymore
        self.assertEqual(s2, None)

    def test_18_add_and_delete_password_reset(self):
        p1 = PasswordReset("recoverycode", "cornelius",
                           "realm",
                           expiration=datetime.now() + timedelta(seconds=120))
        p1.save()
        p2 = PasswordReset.query.filter_by(username="cornelius",
                                           realm="realm").first()
        self.assertTrue(p2.recoverycode, "recoverycode")

    def test_19_add_update_delete_eventhandler(self):
        # Bind the module "usernotice" to the enroll event
        event = "enroll"
        event_update = "init"
        handlermodule = "usernotice"
        action = "email"
        condition = "always"
        options = {"mailserver": "blafoo",
                   "option2": "value2"}
        conditions = {"user_type": "admin"}
        eh1 = EventHandler("ev1", event, handlermodule=handlermodule,
                           action=action, condition=condition,
                           options=options, conditions=conditions)
        self.assertTrue(eh1)

        self.assertEqual(eh1.event, event)
        self.assertEqual(eh1.handlermodule, handlermodule)
        self.assertEqual(eh1.action, action)
        self.assertEqual(eh1.condition, condition)
        self.assertEqual(eh1.options[0].Key, "mailserver")
        self.assertEqual(eh1.options[0].Value, "blafoo")
        self.assertEqual(eh1.options[1].Key, "option2")
        self.assertEqual(eh1.options[1].Value, "value2")
        self.assertEqual(eh1.conditions[0].Key, "user_type")
        self.assertEqual(eh1.conditions[0].Value, "admin")

        eh1_id = eh1.id

        # update eventhandler
        EventHandler("ev1", event_update, handlermodule=handlermodule,
                     action=action, condition=condition,
                     options=options, ordering=0, id=eh1_id)
        self.assertEqual(eh1.event, event_update)

        # Update option value
        EventHandlerOption(eh1_id, Key="mailserver", Value="mailserver")
        self.assertEqual(eh1.options[0].Value, "mailserver")

        # Add Option
        EventHandlerOption(eh1_id, Key="option3", Value="value3")
        self.assertEqual(eh1.options[2].Key, "option3")
        self.assertEqual(eh1.options[2].Value, "value3")

        # Update condition value
        EventHandlerCondition(eh1_id, Key="user_type", Value="user")
        self.assertEqual(eh1.conditions[0].Value, "user")

        # Add condition
        EventHandlerCondition(eh1_id, Key="result_value", Value="True")
        self.assertEqual(eh1.conditions[0].Key, "result_value")
        self.assertEqual(eh1.conditions[0].Value, "True")
        self.assertEqual(eh1.conditions[1].Key, "user_type")
        self.assertEqual(eh1.conditions[1].Value, "user")

        # Delete event handler
        eh1.delete()

    def test_20_add_update_delete_smsgateway(self):
        name = "myGateway"
        provider_module = "privacyidea.lib.smsprovider.httpbla"
        provider_module2 = "module2"
        gw = SMSGateway(name, provider_module, options={"k": "v"})

        self.assertTrue(gw)

        self.assertEqual(gw.identifier, name)
        self.assertEqual(gw.providermodule, provider_module)

        # update SMS gateway, key "k" should not exist anymore!
        SMSGateway(name, provider_module2,
                   options={"k1": "v1"})
        self.assertEqual(gw.providermodule, provider_module2)
        self.assertEqual(gw.options[0].Key, "k1")
        self.assertEqual(gw.options[0].Value, "v1")

        # Delete gateway
        gw.delete()

    def test_21_add_update_delete_clientapp(self):
        # MySQLs DATETIME type supports only seconds so we have to mock now()
        current_time = datetime(2018, 3, 4, 5, 6, 8)
        with mock.patch('privacyidea.models.datetime') as mock_dt:
            mock_dt.now.return_value = current_time

            ClientApplication(ip="1.2.3.4", hostname="host1",
                              clienttype="PAM", node="localnode").save()

        c = ClientApplication.query.filter(ClientApplication.ip == "1.2.3.4").first()
        self.assertEqual(c.hostname, "host1")
        self.assertEqual(c.ip, "1.2.3.4")
        self.assertEqual(c.clienttype, "PAM")
        t1 = c.lastseen

        self.assertIn("localnode", repr(c))

        ClientApplication(ip="1.2.3.4", hostname="host1",
                          clienttype="PAM", node="localnode").save()
        c = ClientApplication.query.filter(ClientApplication.ip == "1.2.3.4").first()
        self.assertGreater(c.lastseen, t1, c)

        ClientApplication.query.filter(ClientApplication.id == c.id).delete()
        c = ClientApplication.query.filter(ClientApplication.ip == "1.2.3.4").first()
        self.assertEqual(c, None)

    def test_22_subscription(self):
        Subscription(application="otrs", for_name="customer",
                     for_email="customer@example.com", for_phone="12345",
                     by_name="provider", by_email="p@example.com",
                     level="Gold").save()
        s = Subscription.query.filter(Subscription.application == "otrs").first()
        self.assertEqual(s.application, "otrs")
        self.assertEqual(s.for_name, "customer")
        self.assertEqual(s.for_email, "customer@example.com")
        self.assertEqual(s.for_phone, "12345")
        self.assertEqual(s.by_name, "provider")
        self.assertEqual(s.by_email, "p@example.com")
        self.assertEqual(s.level, "Gold")

        # Update the entry
        Subscription(application="otrs", for_phone="11111",
                     by_url="https://support.com",
                     signature="1234567890", level="Silver").save()
        s = Subscription.query.filter(
            Subscription.application == "otrs").first()
        self.assertEqual(s.application, "otrs")
        self.assertEqual(s.for_name, "customer")
        self.assertEqual(s.for_email, "customer@example.com")
        self.assertEqual(s.for_phone, "11111")
        self.assertEqual(s.by_name, "provider")
        self.assertEqual(s.by_email, "p@example.com")
        self.assertEqual(s.by_url, "https://support.com")
        self.assertEqual(s.level, "Silver")

        # delete entry
        Subscription.query.filter(Subscription.application == "otrs").delete()
        s = Subscription.query.filter(
            Subscription.application == "otrs").first()
        self.assertEqual(s, None)

    def test_23_usercache(self):
        username = "cornelius"
        resolver = "resolver1"
        user_id = 1
        # we don't need a timestamp with microseconds here, the MySQL DATETIME
        # type doesn't support it out of the box anyway
        timestamp = datetime.now().replace(microsecond=0)

        # create a user in the cache
        cached_user = UserCache(username, username, resolver, user_id, timestamp)
        self.assertTrue(cached_user)
        cached_user.save()

        # search a user in the cache
        find_user = UserCache.query.filter(UserCache.username ==
                                           username).first()
        self.assertTrue(find_user)
        self.assertEqual(find_user.user_id, str(user_id))
        self.assertEqual(find_user.resolver, resolver)
        self.assertEqual(find_user.timestamp, timestamp)  # TODO: Sensible, or might we have a small loss of precision?

        # search the user by his used_login
        # search a user in the cache
        find_user = UserCache.query.filter(UserCache.used_login ==
                                           username).first()
        self.assertTrue(find_user)
        self.assertEqual(find_user.user_id, str(user_id))
        self.assertEqual(find_user.resolver, resolver)

        # delete the user from the cache
        r = find_user.delete()
        self.assertTrue(r)
        find_user = UserCache.query.filter(UserCache.username ==
                                           username).first()
        self.assertFalse(find_user)

    def test_24_add_and_delete_privacyideaserver(self):
        pi1 = PrivacyIDEAServer(identifier="myserver",
                                url="https://pi.example.com")
        pi1.save()
        pi2 = PrivacyIDEAServer.query.filter_by(identifier="myserver").first()
        self.assertEqual(pi2.url, "https://pi.example.com")
        self.assertFalse(pi2.tls)

        # Update the server
        PrivacyIDEAServer(identifier="myserver",
                          url="https://pi2.example.com", tls=True,
                          description="test").save()
        modified_server = PrivacyIDEAServer.query.filter_by(
            identifier="myserver").first()

        self.assertTrue(modified_server.tls, "100.2.3.4")
        self.assertEqual(modified_server.description, "test")
        self.assertEqual(modified_server.url, "https://pi2.example.com")

        # Delete the server
        pi2.delete()
        # Try to find the server
        pi2 = PrivacyIDEAServer.query.filter_by(identifier="myserver").first()
        # The server does not exist anymore
        self.assertEqual(pi2, None)

    def test_25_eventcounter(self):
        counter = EventCounter("test_counter", 10)
        counter.save()
        counter2 = EventCounter.query.filter_by(counter_name="test_counter").first()
        self.assertEqual(counter2.counter_value, 10)
        self.assertEqual(counter2.node, "")

        counter2.increase()
        counter2.increase()

        counter3 = EventCounter.query.filter_by(counter_name="test_counter").first()
        self.assertEqual(counter3.counter_value, 12)

        counter3.decrease()

        counter4 = EventCounter.query.filter_by(counter_name="test_counter").first()
        self.assertEqual(counter4.counter_value, 11)

        counter4.decrease()

        counter5 = EventCounter.query.filter_by(counter_name="test_counter").first()
        self.assertEqual(counter5.counter_value, 10)

        counter6 = EventCounter("test_counter", 4, "othernode")
        self.assertEqual(counter6.counter_value, 4)
        self.assertEqual(counter6.node, "othernode")

        counter_value = db.session.query(func.sum(EventCounter.counter_value))\
            .filter(EventCounter.counter_name == "test_counter").one()[0]
        self.assertEqual(counter_value, 14)

        counters7 = EventCounter.query.filter_by(counter_name="test_counter").all()
        self.assertEqual(len(counters7), 2)

        counter8 = EventCounter.query.filter_by(counter_name="test_counter", node="othernode")
        counter8.delete()

        counters9 = EventCounter.query.filter_by(counter_name="test_counter").all()
        self.assertEqual(len(counters9), 1)
        counters9[0].delete()

        counter10 = EventCounter.query.filter_by(counter_name="test_counter").first()
        self.assertEqual(counter10, None)

    def test_26_periodictask(self):
        current_utc_time = datetime(2018, 3, 4, 5, 6, 8)
        with mock.patch('privacyidea.models.datetime') as mock_dt:
            mock_dt.utcnow.return_value = current_utc_time

            task1 = PeriodicTask("task1", False, "0 5 * * *", ["localhost"], "some.module", 2, {
                "key1": "value2",
                "KEY2": True,
                "key3": "öfføff",
            })
            task2 = PeriodicTask("some other task", True, "0 6 * * *", ["localhost"], "some.other.module", 1, {
                "foo": "bar"
            })

        self.assertEqual(PeriodicTask.query.filter_by(name="task1").one(), task1)
        self.assertEqual(PeriodicTask.query.filter_by(name="some other task").one(), task2)
        self.assertEqual(PeriodicTaskOption.query.filter_by(periodictask_id=task1.id, key="KEY2").one().value,
                         "True")
        # Values are converted to strings
        self.assertEqual(task1.get(), {
            "id": task1.id,
            "name": "task1",
            "active": False,
            "interval": "0 5 * * *",
            # we get a timezone-aware datetime here
            "last_update": current_utc_time.replace(tzinfo=tzutc()),
            "nodes": ["localhost"],
            "taskmodule": "some.module",
            "ordering": 2,
            "options": {
                "key1": "value2",
                "KEY2": "True",
                "key3": "öfføff",
            },
            "retry_if_failed": True,
            "last_runs": {}})

        # register a run
        task1.set_last_run("localhost", datetime(2018, 3, 4, 5, 6, 7))

        # assert we can update the task
        later_utc_time = current_utc_time + timedelta(seconds=1)
        with mock.patch('privacyidea.models.datetime') as mock_dt:
            mock_dt.utcnow.return_value = later_utc_time
            PeriodicTask("task one", True, "0 8 * * *", ["localhost", "otherhost"], "some.module", 3, {
                "KEY2": "value number 2",
                "key 4": 1234
            }, id=task1.id)
        # the first run for otherhost
        task1.set_last_run("otherhost", datetime(2018, 8, 9, 10, 11, 12))
        result = PeriodicTask.query.filter_by(name="task one").one().get()
        self.assertEqual(result,
                         {
                             "id": task1.id,
                             "active": True,
                             "name": "task one",
                             "interval": "0 8 * * *",
                             "last_update": later_utc_time.replace(tzinfo=tzutc()),
                             "nodes": ["localhost", "otherhost"],
                             "taskmodule": "some.module",
                             "ordering": 3,
                             "options": {"KEY2": "value number 2",
                                         "key 4": "1234"},
                             'retry_if_failed': True,
                             "last_runs": {
                                 "localhost": datetime(2018, 3, 4, 5, 6, 7, tzinfo=tzutc()),
                                 "otherhost": datetime(2018, 8, 9, 10, 11, 12, tzinfo=tzutc()),
                             }
                         })
        # assert all old options are removed
        self.assertEqual(PeriodicTaskOption.query.filter_by(periodictask_id=task1.id, key="key3").count(), 0)
        # the second run for localhost
        task1.set_last_run("localhost", datetime(2018, 3, 4, 5, 6, 8))
        result = PeriodicTask.query.filter_by(name="task one").one().get()
        self.assertEqual(result,
                         {
                             "id": task1.id,
                             "active": True,
                             "name": "task one",
                             "interval": "0 8 * * *",
                             "last_update": later_utc_time.replace(tzinfo=tzutc()),
                             "nodes": ["localhost", "otherhost"],
                             "taskmodule": "some.module",
                             "ordering": 3,
                             "options": {"KEY2": "value number 2",
                                         "key 4": "1234"},
                             'retry_if_failed': True,
                             "last_runs": {
                                 "localhost": datetime(2018, 3, 4, 5, 6, 8, tzinfo=tzutc()),
                                 "otherhost": datetime(2018, 8, 9, 10, 11, 12, tzinfo=tzutc()),
                             }
                         })

        # remove "localhost", assert the last run is removed
        PeriodicTask("task one", True, "0 8 * * *", ["otherhost"],
                     "some.module", 4, {"foo": "bar"}, id=task1.id)
        self.assertEqual(PeriodicTaskOption.query.filter_by(periodictask_id=task1.id).count(), 1)
        self.assertEqual(PeriodicTaskLastRun.query.filter_by(periodictask_id=task1.id).one().node, "otherhost")
        # naive timestamp in the database
        self.assertEqual(PeriodicTaskLastRun.query.filter_by(periodictask_id=task1.id).one().timestamp,
                         datetime(2018, 8, 9, 10, 11, 12, tzinfo=None))
        self.assertEqual(PeriodicTaskLastRun.query.filter_by(periodictask_id=task1.id).one().aware_timestamp,
                         datetime(2018, 8, 9, 10, 11, 12, tzinfo=tzutc()))

        # remove the tasks, everything is removed
        task1.delete()
        self.assertEqual(PeriodicTaskOption.query.count(), 1)  # from task2
        self.assertEqual(PeriodicTaskLastRun.query.count(), 0)
        task2.delete()
        self.assertEqual(PeriodicTaskOption.query.count(), 0)

    def test_27_monitoring_stats(self):
        # Simple test to write data to the monitoring stats table
        key1 = "user_count"
        key2 = "successful_auth"
        utcnow = datetime.utcnow()
        MonitoringStats(utcnow - timedelta(seconds=1), key1, 15).save()
        MonitoringStats(utcnow, key1, 21).save()
        MonitoringStats(utcnow, key2, 123).save()

        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 2)
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key2).count(), 1)

        # Delete all entries
        MonitoringStats.query.delete()
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 0)
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key2).count(), 0)


class TokenModelTestCaseDeleting(MyTestCase):

    def test_01_create_and_delete_resolver(self):
        r = Resolver("try_delete", "passwdresolver")
        rid = r.save()
        self.assertTrue(r.name is not None, r.name)
        self.assertTrue(r.rtype is not None, r.rtype)
        # Add configuration to the resolver
        conf = ResolverConfig(r.id, "fileName", "this_very_specific_file_try_delete")
        conf.save()

        # Read Resolver
        r1 = Resolver.query.filter_by(name="try_delete").first()
        self.assertTrue(r1.rtype == "passwdresolver", r1.rtype)
        self.assertEqual(rid, r1.id)

        # Get the option
        rc = ResolverConfig.query.filter_by(Value="this_very_specific_file_try_delete").first()
        self.assertTrue(rc)
        self.assertEqual(rc.resolver_id, rid)

        # Now delete the Resolver and check, if there is no config left
        reso = Resolver.query.filter_by(name="try_delete").first()
        reso.delete()
        rc = ResolverConfig.query.filter_by(Value="this_very_specific_file_try_delete").first()
        self.assertIsNone(rc)


class TokengroupTestCase(MyTestCase):

    def test_01_create_update_delete_tokengroup(self):
        tg = Tokengroup("gruppe1", "coolest group ever")
        self.assertIsInstance(tg, Tokengroup)
        rid = tg.save()
        self.assertGreaterEqual(rid, 1)

        # Test, if it does exist
        r = Tokengroup.query.filter_by(name="gruppe1").all()
        self.assertEqual(len(r), 1)
        self.assertIsInstance(r[0], Tokengroup)

        # delete
        r = tg.delete()
        self.assertEqual(r, rid)

        # Test, if it is gone
        r = Tokengroup.query.filter_by(name="gruppe1").all()
        self.assertEqual(len(r), 0)

    def test_02_assign_tokens_to_tokengroup(self):
        # create token groups
        tg1 = Tokengroup("gruppe1", "coolest group ever")
        self.assertIsInstance(tg1, Tokengroup)
        rid = tg1.save()
        self.assertGreaterEqual(rid, 1)
        tg2 = Tokengroup("gruppe2", "2nd group")
        self.assertIsInstance(tg2, Tokengroup)
        rid = tg2.save()
        self.assertGreaterEqual(rid, 1)

        # create tokens
        tok1 = Token(tokentype="spass", serial="tok1")
        tok1.save()
        tok2 = Token(tokentype="spass", serial="tok2")
        tok2.save()
        self.assertEqual(tok1.serial, "tok1")
        self.assertEqual(tok2.serial, "tok2")

        # assign tokens to token groups
        t = TokenTokengroup(token_id=tok1.id, tokengroupname="gruppe1").save()
        t = TokenTokengroup(token_id=tok1.id, tokengroupname="gruppe2").save()
        t = TokenTokengroup(token_id=tok2.id, tokengroup_id=tg2.id).save()
        ttg = TokenTokengroup.query.all()
        self.assertEqual(len(ttg), 3)
        # It does not change anything, if we try to save the same assignment!
        TokenTokengroup(token_id=tok2.id, tokengroup_id=tg2.id).save()
        ttg = TokenTokengroup.query.all()
        self.assertEqual(len(ttg), 3)

        ttg = TokenTokengroup.query.filter_by(token_id=tok1.id).all()
        self.assertEqual(len(ttg), 2)

        ttg = TokenTokengroup.query.filter_by(token_id=tok2.id).all()
        self.assertEqual(len(ttg), 1)

        self.assertEqual(len(tok1.tokengroup_list), 2)
        self.assertEqual(len(tok2.tokengroup_list), 1)

        self.assertEqual(tok2.tokengroup_list[0].tokengroup.name, "gruppe2")

        # remove tokengroups and check that tokentokengroups assignments are removed
        tg1.delete()
        ttg = TokenTokengroup.query.all()
        self.assertEqual(len(ttg), 2)

        tg2.delete()
        ttg = TokenTokengroup.query.all()
        self.assertEqual(len(ttg), 0)

        # cleanup
        tok1.delete()
        tok2.delete()


class ServiceidTestCase(MyTestCase):

    def test_01_create_update_delete_servicid(self):
        si = Serviceid("webserver", "The group of all our webservers")
        self.assertIsInstance(si, Serviceid)
        rid = si.save()
        self.assertGreaterEqual(rid, 1)

        # Test, if it does exist
        r = Serviceid.query.filter_by(name="webserver").all()
        self.assertEqual(len(r), 1)
        self.assertIsInstance(r[0], Serviceid)
        self.assertEqual("The group of all our webservers", r[0].Description)

        # update the description
        r[0].Description = "New description"
        r[0].save()

        # Check it:
        r = Serviceid.query.filter_by(name="webserver").all()
        self.assertEqual(len(r), 1)
        self.assertIsInstance(r[0], Serviceid)
        self.assertEqual("New description", r[0].Description)

        # delete
        r = si.delete()
        self.assertEqual(r, rid)

        # Test, if it is gone
        r = Serviceid.query.filter_by(name="webserver").all()
        self.assertEqual(len(r), 0)


class ResolverRealmTestCase(MyTestCase):

    def test_01_resolver_realm_with_nodes(self):
        nd1_uuid = "8e4272a9-9037-40df-8aa3-976e4a04b5a9"
        nd2_uuid = "d1d7fde6-330f-4c12-88f3-58a1752594bf"
        NodeName(id=nd1_uuid, name="Node1").save()
        NodeName(id=nd2_uuid, name="Node2").save()

        res = Resolver("resolver1", "passwdresolver")
        res.save()
        # Add configuration to the resolver
        ResolverConfig(res.id, "fileName", "tests/testdata/passwd").save()

        re1 = Realm("realm1")
        re1.save()

        # Put the resolver into the realm by name
        ResolverRealm(resolver_name="resolver1",
                      realm_name="realm1").save()

        self.assertIn(res.id, [x.resolver.id for x in re1.resolver_list])
        self.assertEqual(re1.resolver_list[0].node_uuid, "", re1.resolver_list[0])
        re1.delete()

        re2 = Realm("realm2")
        re2.save()
        ResolverRealm(resolver_name="resolver1",
                      realm_name="realm2", node_uuid=nd1_uuid).save()
        self.assertIn(res.id, [x.resolver.id for x in re2.resolver_list])
        self.assertEqual(re2.resolver_list[0].node_uuid, nd1_uuid, re1.resolver_list[0])
        re2.delete()

    # TODO: add resolver realm config with ids and different nodes
    # TODO: same nodes with different timestamps



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
import json
from dataclasses import dataclass
from unittest.mock import patch

from webauthn import base64url_to_bytes
from webauthn.helpers import bytes_to_base64url
from webauthn.helpers.cose import COSEAlgorithmIdentifier
from webauthn.helpers.structs import AttestationConveyancePreference

from privacyidea.lib.challenge import get_challenges
from privacyidea.lib.error import EnrollmentError, ParameterError, ResourceNotFoundError
from privacyidea.lib.policy import SCOPE, ACTION
from privacyidea.lib.token import (init_token, remove_token, unassign_token)
from privacyidea.lib.fido2.util import get_credential_ids_for_user, get_fido2_token_by_credential_id, hash_credential_id
from privacyidea.lib.fido2.challenge import create_fido2_challenge, verify_fido2_challenge
from privacyidea.lib.tokenclass import ROLLOUTSTATE, TokenClass
from privacyidea.lib.tokens.passkeytoken import PasskeyTokenClass
from privacyidea.lib.fido2.policy_action import FIDO2PolicyAction, PasskeyAction
from privacyidea.lib.user import User
from privacyidea.models import TokenCredentialIdHash
from tests.base import MyTestCase
from tests.passkey_base import PasskeyTestBase


class PasskeyTokenTestCase(PasskeyTestBase, MyTestCase):
    """
    The tests in this class are meant to be run consecutively, because they build on each other.
    """

    def setUp(self):
        PasskeyTestBase.setUp(self)
        self.setUp_user_realms()
        self.user = User(login="hans", realm=self.realm1, resolver=self.resolvername1)

    def _create_token(self) -> TokenClass:
        registration_request = self._initialize_registration()
        token = registration_request.token
        token.update(registration_request.registration_response)
        self.assertEqual(token.token.rollout_state, ROLLOUTSTATE.ENROLLED)
        return token

    @dataclass(frozen=True)
    class RegistrationRequest:
        token: TokenClass
        init_detail: dict
        registration_response: dict

    def _initialize_registration(self, get_init_detail_options: dict = None) -> RegistrationRequest:
        if get_init_detail_options is None:
            get_init_detail_options = {}
        token = init_token({"type": "passkey"}, user=self.user)
        self.assertIsInstance(token, PasskeyTokenClass)
        self.assertEqual(token.rollout_state, ROLLOUTSTATE.CLIENTWAIT)
        param = {}
        if get_init_detail_options:
            param.update(get_init_detail_options)
        # Overwrite these values if they have been in options
        param.update({
            FIDO2PolicyAction.RELYING_PARTY_ID: self.rp_id,
            FIDO2PolicyAction.RELYING_PARTY_NAME: self.rp_id
        })
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.registration_challenge
            init_detail = token.get_init_detail(param)
        self.assertIn("serial", init_detail)
        self.assertIn("passkey_registration", init_detail)
        self.assertIn("transaction_id", init_detail)
        registration_response = {
            "attestationObject": self.registration_attestation,
            "clientDataJSON": self.registration_client_data,
            "credential_id": self.credential_id,
            "rawId": self.credential_id,
            "authenticatorAttachment": self.authenticator_attachment,
            "HTTP_ORIGIN": self.expected_origin,
            FIDO2PolicyAction.RELYING_PARTY_ID: self.rp_id,
            "transaction_id": init_detail["transaction_id"]
        }

        return self.RegistrationRequest(token, init_detail, registration_response)

    def _initialize_authentication(self) -> dict:
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.authentication_challenge_no_uv
            challenge = create_fido2_challenge(self.rp_id)
            self.assertEqual(challenge["challenge"], self.authentication_challenge_no_uv)
            return challenge

    def test_01_class_values(self):
        class_prefix = PasskeyTokenClass.get_class_prefix()
        self.assertEqual(class_prefix, "PIPK")

        class_type = PasskeyTokenClass.get_class_type()
        self.assertEqual(class_type, "passkey")

        class_info = PasskeyTokenClass.get_class_info()
        self.assertEqual(class_info["type"], "passkey")
        self.assertIn("title", class_info)
        self.assertIn("description", class_info)
        self.assertIn("user", class_info)
        self.assertEqual(class_info["user"], ["enroll"])
        self.assertIn("ui_enroll", class_info)
        self.assertEqual(class_info["ui_enroll"], ["admin", "user"])

        self.assertIn("policy", class_info)
        policy = class_info["policy"]
        self.assertIn(SCOPE.AUTH, policy)
        policy_auth = policy[SCOPE.AUTH]
        self.assertIn(ACTION.CHALLENGETEXT, policy_auth)
        self.assertIn(SCOPE.ENROLL, policy)
        policy_enroll = policy[SCOPE.ENROLL]
        self.assertIn(PasskeyAction.AttestationConveyancePreference, policy_enroll)

    def test_02_init_defaults(self):
        """
        Test the basic enrollment with just the settings that are required to make it work. The token is persisted for
        the next test.
        """
        registration_request = self._initialize_registration()
        token = registration_request.token
        init_detail = registration_request.init_detail
        self.assertTrue(token.get_serial())
        self.first_serial = token.get_serial()

        self.validate_default_passkey_registration(init_detail["passkey_registration"])
        passkey_registration = init_detail["passkey_registration"]

        self.assertEqual(len(passkey_registration["pubKeyCredParams"]), 2)
        self.assertEqual(passkey_registration["pubKeyCredParams"][0]["type"], "public-key")
        self.assertEqual(passkey_registration["pubKeyCredParams"][1]["type"], "public-key")
        self.assertIn(passkey_registration["pubKeyCredParams"][0]["alg"], [-7, -257])
        self.assertIn(passkey_registration["pubKeyCredParams"][1]["alg"], [-7, -257])
        # ExcludeCredentials should be empty because no other passkey token is registered for the user
        self.assertEqual(len(passkey_registration["excludeCredentials"]), 0)

        # Complete the registration
        update_response = token.update(registration_request.registration_response)
        self.assertTrue(update_response)
        self.assertIn("details", update_response)
        self.assertEqual(update_response["details"]["serial"], self.first_serial)
        # Verify that the token is enrolled and token info was written
        self.assertEqual(token.rollout_state, ROLLOUTSTATE.ENROLLED)
        token_info = token.get_tokeninfo()
        self.assertTrue(token_info["credential_id_hash"])
        self.assertTrue(token_info["public_key"])
        self.assertTrue(token_info["aaguid"])
        self.assertTrue(token_info["sign_count"])
        self.assertTrue(token_info["fido2_user_id"])
        # Since attestation was passed, the cert should be present and the description should be set from the cert CN
        self.assertTrue(token_info["attestation_certificate"])
        self.assertEqual(token.token.description, "Yubico U2F EE Serial 2109467376")
        # The token key is the credential id
        self.assertEqual(token.token.get_otpkey().getKey().decode('utf-8'), self.credential_id)
        # Verify that the credential_id hash has been written to the database
        tcih = TokenCredentialIdHash.query.filter(
            TokenCredentialIdHash.credential_id_hash == token_info["credential_id_hash"]).one()
        self.assertEqual(tcih.token_id, token.token.id)
        # Kep the token for the next test

    def test_03_init_custom_settings(self):
        """
        Test enrollment with all custom settings. The settings that are not changeable are not checked again.
        This also just tests the registration options generated by the server, because in the previous test, the
        registration was done with attestation.
        This basically tests that the policies for enrollment are loaded by the prepolicy 'fido2_enroll', that they work
        and that excludeCredentials is set correctly.
        """
        # UserVerification is ignored by most browsers when residentKey is set to required
        # However, it can still be configured and is tested here
        # The Key Algorithm are all that are selectable in the policy via
        # privacyidea.lib.tokens.webauthntoken.PUBLIC_KEY_CREDENTIAL_ALGORITHMS
        key_algorithms = [COSEAlgorithmIdentifier.ECDSA_SHA_256,
                          COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_256,
                          COSEAlgorithmIdentifier.RSASSA_PSS_SHA_256]
        # Because of how token are created currently (API level calls functions on the token object directly),
        # this has to be done here instead of on the lib level
        # There should be the credential_id of the token created in the previous test
        registered_credential_ids = get_credential_ids_for_user(self.user)
        self.assertEqual(len(registered_credential_ids), 1)

        options = {
            FIDO2PolicyAction.PUBLIC_KEY_CREDENTIAL_ALGORITHMS: key_algorithms,
            PasskeyAction.AttestationConveyancePreference: AttestationConveyancePreference.ENTERPRISE,
            FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT: "required",
            "registered_credential_ids": registered_credential_ids
        }
        registration_request = self._initialize_registration(options)
        token = registration_request.token
        init_detail = registration_request.init_detail
        passkey_registration = init_detail["passkey_registration"]
        # Verify that the custom settings are set in the registration options
        self.assertIn("pubKeyCredParams", passkey_registration)
        self.assertEqual(len(passkey_registration["pubKeyCredParams"]), 3)
        for element in passkey_registration["pubKeyCredParams"]:
            self.assertEqual(element["type"], "public-key")
            self.assertIn(element["alg"], key_algorithms)
        self.assertEqual(passkey_registration["attestation"], "enterprise")
        self.assertEqual(passkey_registration["authenticatorSelection"]["userVerification"], "required")
        # excludeCredentials have the format [{'id': 'credential_id', 'type': 'public-key'}, {...}, ...]
        # The type is always 'public-key'
        self.assertEqual(passkey_registration["excludeCredentials"][0]["id"], registered_credential_ids[0])
        remove_token(serial=token.get_serial())
        # Remove the token from the previous test
        token = get_fido2_token_by_credential_id(registered_credential_ids[0])
        remove_token(serial=token.get_serial())
        # Verify that the hashed credential has been removed
        credential_id_hash = hash_credential_id(registered_credential_ids[0])
        tcih = TokenCredentialIdHash.query.filter(
            TokenCredentialIdHash.credential_id_hash == credential_id_hash).first()
        self.assertFalse(tcih)

    def test_04_init_fail_wrong_data(self):
        registration_request = self._initialize_registration()
        token = registration_request.token

        # Complete the registration with tampered data
        with self.assertRaises(EnrollmentError):
            registration_response = registration_request.registration_response
            registration_response["attestationObject"] = self.registration_attestation.replace("e", "f")
            token.update(registration_response)

        # Change the origin, should be reflected in the error message
        client_data = base64url_to_bytes(self.registration_client_data).decode('utf-8')
        client_data_json = json.loads(client_data)
        self.assertTrue(client_data_json["origin"])
        client_data_json["origin"] = "https://evil.nils:5000"
        tampered_client_data = bytes_to_base64url(json.dumps(client_data_json).encode('utf-8'))
        registration_response["clientDataJSON"] = tampered_client_data
        with self.assertRaises(EnrollmentError) as ex:
            token.update(registration_response)
        self.assertIn('Unexpected client data origin "https://evil.nils:5000",'
                      ' expected "https://cool.nils:5000"',
                      ex.exception.message)
        # Invalid client data (json format)
        with self.assertRaises(EnrollmentError) as ex:
            registration_response["clientDataJSON"] = self.registration_client_data.replace("a", "b")
            token.update(registration_response)
        self.assertIn("Invalid JSON structure", ex.exception.message)

    def test_05_init_exceptions(self):
        # Also test setting a description in the first step
        token: TokenClass = init_token({"type": "passkey", "description": "test"}, user=self.user)
        self.assertEqual(token.token.description, "test")
        param = {FIDO2PolicyAction.RELYING_PARTY_ID: self.rp_id,
                 FIDO2PolicyAction.RELYING_PARTY_NAME: self.rp_id}

        # Remove the user before the registration
        unassign_token(token.get_serial())
        with self.assertRaises(ParameterError) as ex:
            token.get_init_detail(param)
        self.assertIn("User must be provided", ex.exception.message)

        # Invalid enrollment state returns empty init detail
        token.token.rollout_state = ROLLOUTSTATE.PENDING
        init_detail = token.get_init_detail(param)
        self.assertFalse(init_detail)
        remove_token(serial=token.get_serial())

        # Registration challenge expired/deleted
        registration_request = self._initialize_registration()
        challenges = get_challenges(registration_request.token.get_serial())
        self.assertEqual(len(challenges), 1)
        challenges[0].delete()
        with self.assertRaises(EnrollmentError) as ex:
            registration_request.token.update(registration_request.registration_response)
        self.assertIn("The enrollment challenge does not exist", ex.exception.message)
        remove_token(serial=registration_request.token.get_serial())

    def test_06_authenticate_success(self):
        token = self._create_token()
        challenge = self._initialize_authentication()
        # UserVerification is preferred by default
        authentication_response = self.authentication_response_no_uv
        authentication_response["HTTP_ORIGIN"] = self.expected_origin
        success = verify_fido2_challenge(challenge["transaction_id"], token, authentication_response)
        self.assertEqual(success, 1)
        remove_token(serial=token.get_serial())

    def test_07_authenticate_fails(self):
        token = self._create_token()

        # Challenge does not exist
        with self.assertRaises(ResourceNotFoundError) as ex:
            verify_fido2_challenge("10039292755795086078", token, {})

        # Wrong signature
        challenge = self._initialize_authentication()
        authentication_response = self.authentication_response_no_uv
        authentication_response["HTTP_ORIGIN"] = self.expected_origin
        authentication_response["signature"] = authentication_response["signature"].replace("a", "b")
        success = verify_fido2_challenge(challenge["transaction_id"], token, authentication_response)
        self.assertEqual(success, -1)
        remove_token(serial=token.get_serial())

    def test_08_no_tokencredentialidhash_entry(self):
        _ = self._create_token()
        # Remove the tokencredentialidhash entry
        credential_id_hash = hash_credential_id(self.credential_id)
        tcih = TokenCredentialIdHash.query.filter(
            TokenCredentialIdHash.credential_id_hash == credential_id_hash).one()
        self.assertTrue(tcih)
        tcih.delete()
        # Try to get the token by credential id
        token = get_fido2_token_by_credential_id(self.credential_id)
        # Check that the credential id hash has been added again
        tcih = TokenCredentialIdHash.query.filter(
            TokenCredentialIdHash.credential_id_hash == credential_id_hash).one()
        self.assertTrue(tcih)

    def test_09_duplicate_tokencredentialidhash_entry(self):
        token1 = self._create_token()
        # Get the tokencredentialidhash entry
        credential_id_hash = hash_credential_id(self.credential_id)
        tcih = TokenCredentialIdHash.query.filter(
            TokenCredentialIdHash.credential_id_hash == credential_id_hash).one()
        self.assertTrue(tcih)
        self.assertEqual(tcih.token_id, token1.token.id)
        # Create a new token with the same credential id, which will overwrite the existing TCIH entry
        token2 = self._create_token()
        tcih = TokenCredentialIdHash.query.filter(
            TokenCredentialIdHash.credential_id_hash == credential_id_hash).one()
        self.assertTrue(tcih)
        self.assertEqual(tcih.token_id, token2.token.id)



"""add column position for event handlers

Revision ID: 1a0710df148b
Revises: 19f727d285e2
Create Date: 2018-08-03 12:36:17.091876

"""

# revision identifiers, used by Alembic.
revision = '1a0710df148b'
down_revision = '19f727d285e2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.add_column('eventhandler', sa.Column('position', sa.Unicode(length=10), default=u"post"))
    except Exception as exx:
        print("position column in eventhandler table obviously already exists.")
        print(exx)


def downgrade():
    op.drop_column('eventhandler', 'position')
